#include "cell.h"
#include <vector>
#include "rom_list.h"
#include <math.h>
#include <sys/paths.h>
#include "ps3video.h"
#include "ps3audio.h"
#include "cellinput.h"
#include "ini.h"
#include <unistd.h> 
#include <pthread.h>

PS3Graphics* Graphics;
CellInputFacade* PS3input  = 0;

#define VIDEO_WIDTH  320 
#define VIDEO_HEIGHT 240
#define SAMPLERATE_48KHZ 48000

static pthread_t thread;

bool is_running = 0;

SYS_PROCESS_PARAM(1001, 0x10000);

static bool runbios = 0;
char rom_path[256];
FileIniConfig Iniconfig;

enum
{
	IN_MENU,
	IN_GAME
};

#define SB_SIZE 3840
unsigned char soundbuffer[3840];
int state      = 0;
int EmuRun     = 0;

void InitPS3()
{
	Graphics = new PS3Graphics();
	Graphics->Init();

	// FIXME: Is this necessary?
	if (Graphics->InitCg() != CELL_OK)
	{
		printf("Failed to InitCg: %d\n", __LINE__);
		exit(0);
	}

	PS3input = new CellInputFacade();
	Graphics->InitDbgFont();
}

extern "C"
{
#include "shared.h"


static int config_load(void)
{
  /* open configuration file */
  char fname[MAXPATHLEN];
  sprintf (fname, "/dev_hdd0/game/GENP00001/config.bin");
  FILE *fp = fopen(fname, "rb");
  if (fp)
  {
    /* read version */
    char version[16];
    fread(version, 16, 1, fp); 
    fclose(fp);
    if (strncmp(version,CONFIG_VERSION,16))
      return 0;

    /* read file */
    fp = fopen(fname, "rb");
    if (fp)
    {
      fread(&config, sizeof(config), 1, fp);
      fclose(fp);
      return 1;
    }
  }
  return 0;
}

void config_save(void)
{
  /* open configuration file */
  char fname[MAXPATHLEN];
  sprintf (fname, "/dev_hdd0/game/GENP00001/config.bin");
  FILE *fp = fopen(fname, "wb");
  if (fp)
  {
    /* write file */
    fwrite(&config, sizeof(config), 1, fp);
    fclose(fp);
  }
}

void config_default()
{
  /* version TAG */
  strncpy(config.version,CONFIG_VERSION,16);

  int i;

  /* sound options */
  config.psg_preamp     = 150;
  config.fm_preamp      = 100;
  config.hq_fm          = 1;
  config.psgBoostNoise  = 0;
  config.filter         = 1;
  config.low_freq       = 200;
  config.high_freq      = 8000;
  config.lg             = 1.0;
  config.mg             = 1.0;
  config.hg             = 1.0;
  config.lp_range       = 60;
  config.rolloff        = 0.995;
  config.dac_bits       = 14;

  /* system options */
  config.region_detect  = 0;
  config.force_dtack    = 0;
  config.addr_error     = 1;
  config.bios_enabled   = 0;
  config.lock_on        = Settings.ExtraCart;
  config.romtype        = 0;
  config.hot_swap       = 0;

  /* video options */
  config.xshift   = 0;
  config.yshift   = 0;
  config.xscale   = 0;
  config.yscale   = 0;
  config.aspect   = 1;
  config.overscan = 0;

  /* controllers options */
  input.system[0]       = SYSTEM_GAMEPAD;
  input.system[1]       = SYSTEM_GAMEPAD;
  config.gun_cursor[0]  = 1;
  config.gun_cursor[1]  = 1;
  config.invert_mouse   = 0;
  for (i=0;i<MAX_INPUTS;i++)
  {
    if(Settings.PAD)
	config.input[i].padtype = DEVICE_6BUTTON;
	else
    config.input[i].padtype = DEVICE_3BUTTON;
  }
}

/***************************************************************************
 * Genesis Plus Virtual Machine
 *
 ***************************************************************************/
static void load_bios(void)
{
  char* bios;
  /* clear BIOS detection flag */
  config.bios_enabled &= ~2;

  /* open BIOS file */
  FILE *fp = fopen(Iniconfig.biospath, "rb");
  if (fp == NULL) return;

  /* read file */
  fread(bios_rom, 1, 0x800, fp);
  fclose(fp);

  /* check ROM file */
  if (!strncmp((char *)(bios_rom + 0x120),"GENESIS OS", 10))
  {
    /* valid BIOS detected */
    config.bios_enabled |= 2;
  }
}


int Megadrive_INIT()
{
	float GameHz;
	CellVideoOutState VideoSync = Graphics->GetVideoOutState(); 

	/* allocate cart.rom here (10 MBytes) */
	if(!EmuRun){
	cart.rom = (unsigned char*)memalign(32, MAXROMSIZE);


	config_default();

	/* BIOS support */
	load_bios();
	}

	/* Load game */
    if(!load_rom(rom_path))
	{
		printf("Error loading file `%s'.\n",rom_path);
	}

  /* allocate global work bitmap */
	if(!EmuRun){
	memset (&bitmap, 0, sizeof (bitmap));
	bitmap.width  = 320;
	bitmap.height = 224;
    bitmap.depth  = 32;
    bitmap.granularity = 4;
    bitmap.pitch = (bitmap.width * bitmap.granularity);
    bitmap.data   = (unsigned char *)memalign(1024,bitmap.width*bitmap.height*bitmap.granularity);
	bitmap.viewport.w = 320;
	bitmap.viewport.h = 224;
	bitmap.viewport.x = 0;
	bitmap.viewport.y = 0;
    bitmap.remap = 1;

	/* force video update */
	bitmap.viewport.changed = 3;


	printf("Refresh Rate 0x%X \n",VideoSync.displayMode.refreshRates);
	//for fixing the audio speed because the Vsync fuck our timing !!!!
	if(VideoSync.displayMode.refreshRates == CELL_VIDEO_OUT_REFRESH_RATE_59_94HZ)
		GameHz = 60.0;
	else if(VideoSync.displayMode.refreshRates == CELL_VIDEO_OUT_REFRESH_RATE_60HZ)
		GameHz = 60.0;
	else if(VideoSync.displayMode.refreshRates == CELL_VIDEO_OUT_REFRESH_RATE_50HZ)
		GameHz = 50.0;
	else exit(0);

	/* initialize emulation */
	audio_init(SAMPLERATE_48KHZ, GameHz);
	
	system_init();

    system_reset();
	}
	
	EmuRun = 1;

	state = IN_GAME;

	return 0;
}

void ChangePathAndExt(char *file,char* ext,char *out)
{
 int nLen,pos; 
 char *myfiles;
 // Get Size !!! 
 nLen = strlen (file); 
 pos = nLen - 1;

 if ((nLen > 0) && (nLen < 256))
 { 
    //extract file name only
	while (pos >= 0)
	{

		if (file[pos] == '/')
		{

			pos++;
			myfiles = (char*)file + pos;
			printf("file=%s\n", myfiles);
			break;

		}
		pos--;
	}
 
      while (nLen >= 0) { 
 
           // Check for extension character remove extention !!! 
           if (myfiles [nLen] == '.') { 
 
                myfiles [nLen] = '\0'; 
				printf("file no ext =%s\n", myfiles);
                break; 
           } 
 
           nLen --; 
    } 
  // Create output file name and with new extension 
  sprintf (out, "%s/%s%s",out,myfiles,ext);
  printf("Final path %s \n",out);
  } 
}

void LoadState()
{
		printf("save path %s \n",Iniconfig.savpath);
        FILE *f = fopen(Iniconfig.savpath,"rb");
        if (f)
        {
		  printf("enter in loadstate");
          uint8 *buf =(uint8 *)memalign(32,STATE_SIZE);
          fread(buf, STATE_SIZE, 1, f);
          state_load(buf);
          fclose(f);
		  free(buf);
		  printf("out from loadstate");
        }
}

void SaveState()
{
		printf("save path %s \n",Iniconfig.savpath);
        FILE *f = fopen(Iniconfig.savpath,"wb");
        if (f)
        {
		  printf("enter in savestate");
          uint8 *buf =(uint8 *)memalign(32,STATE_SIZE);
          int state_size = state_save(buf);
		  printf("write savestate");
          fwrite(buf, state_size, 1, f);
		  printf("write savestate done!");
          fclose(f);
		  free(buf);
		  printf("out from savestate");
        }
}

int ps3_update_input(void)
{
	uint32_t buttons;
	int i;

	for(i=0;i<2;i++)
	{
	PS3input->UpdateDevice(i);

      /* clear key status */
    input.pad[i] = 0;

    if (PS3input->IsButtonPressed(i, CTRL_UP))	  input.pad[i] |= INPUT_UP;
	else if (PS3input->IsButtonPressed(i,CTRL_DOWN))  input.pad[i] |= INPUT_DOWN;
    if (PS3input->IsButtonPressed(i,CTRL_LEFT))  input.pad[i] |= INPUT_LEFT;
    else if (PS3input->IsButtonPressed(i,CTRL_RIGHT)) input.pad[i] |= INPUT_RIGHT;
    if (PS3input->IsButtonPressed(i,CTRL_SQUARE))   input.pad[i] |= INPUT_A;
	if (PS3input->IsButtonPressed(i,CTRL_CROSS))    input.pad[i] |= INPUT_B;
    if (PS3input->IsButtonPressed(i,CTRL_CIRCLE))   input.pad[i] |= INPUT_C;
    if (PS3input->IsButtonPressed(i,CTRL_START))   input.pad[i] |= INPUT_START;
    if (PS3input->IsButtonPressed(i,CTRL_TRIANGLE))  input.pad[i] |= INPUT_X;
	if (PS3input->IsButtonPressed(i,CTRL_L1))      input.pad[i] |= INPUT_Y;
	if (PS3input->IsButtonPressed(i,CTRL_R1))      input.pad[i] |= INPUT_Z;
    if (PS3input->IsButtonPressed(i,CTRL_SELECT))  input.pad[i] |= INPUT_MODE;
	if ((PS3input->IsButtonPressed(i,CTRL_L2) || PS3input->IsButtonPressed(i,CTRL_R2)))
		return 0;//state = IN_MENU;
	}

	return 1;
}
//end extern C
}

void sysutil_callback (uint64_t status, uint64_t param, void *userdata) {
	(void) param;
	(void) userdata;

	switch (status) {
		case CELL_SYSUTIL_REQUEST_EXITGAME:
			printf("exit from game\n");
			is_running = 0;
			break;
		case CELL_SYSUTIL_DRAWING_BEGIN:
		case CELL_SYSUTIL_DRAWING_END:
			break;
	}
}

static void *SoundThread(void *arg)
{
		int size = audio_update() * 2;
		while (cellAudioPortWriteAvail() > size)
        {
			cellAudioPortWrite((const int16_t* )soundbuffer,size);
        }
}

void RomBrowser()
{
	printf("aspec ration  0x%X \n",(int)Graphics->GetDeviceAspectRatio());
	//detection 16/9 or 4/3 Anonymous
	if((int)Graphics->GetDeviceAspectRatio() == 0x1) //0x1 == 16:9 
		Graphics->SetAspectRatio(0); // 16:9
	else
		Graphics->SetAspectRatio(1); // 4:3

	//browser with roms folder
	MenuMainLoop(Iniconfig.rompath);

	Megadrive_INIT();

	ChangePathAndExt(rom_path,".sav",Iniconfig.savpath);
	
	//clear screen to black
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

}

void StateLoop()
{
	//our main loop for the save state
	PS3input->UpdateDevice(0);
	if ((PS3input->IsButtonPressed(0,CTRL_L3)))    
		SaveState();
	if ((PS3input->IsButtonPressed(0,CTRL_R3)))    
		LoadState();
}
//we parse our ini files
static int handler(void* user, const char* section, const char* name,
                   const char* value)
{
    FileIniConfig* pconfig = (FileIniConfig*)user;

    #define MATCH(s, n) strcmp(section, s) == 0 && strcmp(name, n) == 0
    if (MATCH("genesisplus", "version")) {
        pconfig->version = strdup(value);
    } else if (MATCH("genplusrom", "rompath")) {
        pconfig->rompath = strdup(value);
    } else if (MATCH("genplussav", "savpath")) {
        pconfig->savpath = strdup(value);
    }else if (MATCH("genplussram", "srampath")) {
        pconfig->sram_path = strdup(value);
    }else if (MATCH("genpluscart", "cartpath")) {
        pconfig->cartpath = strdup(value);
    }else if (MATCH("genplusbios", "biospath")) {
        pconfig->biospath = strdup(value);
    }
}

void CreatFolder(char* folders)
{
	if(mkdir(folders,0777))
	{
	gl_dprintf(0.09f,0.05f,FontSize(),"Error folder cannot be created %s !!\nplease check your GenesisConf.ini\n",folders);
	sys_timer_sleep(5);
	sys_process_exit(0);
	}
}

int main() {
    int i;
	struct stat st;

	sys_spu_initialize(6, 1); 
	cellSysutilRegisterCallback(0, (CellSysutilCallback)sysutil_callback, NULL); 

	cellSysmoduleLoadModule(CELL_SYSMODULE_FS);
	cellSysmoduleLoadModule(CELL_SYSMODULE_IO);

	state = IN_MENU;

	is_running = 1;

	InitPS3();
	cellAudioPortInit(SAMPLERATE_48KHZ,SB_SIZE);
	PS3input->Init();

	//read the ini now 
	if (ini_parse("/dev_hdd0/game/GENP00001/GenesisConf.ini", handler, &Iniconfig) < 0)
	{
		gl_dprintf(0.09f,0.05f,FontSize(),"Can't load /dev_hdd0/game/GENP00001/GenesisConf.ini\n");
		sys_timer_sleep(5);
		gl_dprintf(0.09f,0.05f,FontSize(),"Wtf where is the ini!!!!!!!!bye bye try again\n");
		sys_timer_sleep(5);
		sys_process_exit(0);
	}

	printf(" version  %s \n",Iniconfig.version);
	printf(" rompath  %s \n",Iniconfig.rompath);
	printf(" savpath  %s \n",Iniconfig.savpath);
	printf(" srampath %s \n",Iniconfig.sram_path);
	printf(" cartpath %s \n",Iniconfig.cartpath);
	printf(" biospath %s \n",Iniconfig.biospath);

	//Set Extra Cart path
	sprintf(GG_ROM,"%s/ggenie.bin",Iniconfig.cartpath);
	sprintf(AR_ROM,"%s/areplay.bin",Iniconfig.cartpath);
	sprintf(SK_ROM,"%s/sk.bin",Iniconfig.cartpath);
	sprintf(SK_UPMEM,"%s/sk2chip.bin",Iniconfig.cartpath);

	//main path Check if not present creat all folder and exit
	if(stat(Iniconfig.rompath,&st) != 0)
	{
	gl_dprintf(0.09f,0.05f,FontSize(),"Creat generic folder Tree on Genesisplus\n");
	sys_timer_sleep(5);
	CreatFolder(Iniconfig.rompath);
	CreatFolder(Iniconfig.savpath);
	CreatFolder(Iniconfig.cartpath);
	CreatFolder(Iniconfig.sram_path);
	CreatFolder(Iniconfig.biospath);
	gl_dprintf(0.09f,0.05f,FontSize(),"generic folder Tree done!! reboot\nplease put all your roms inside %s\n",Iniconfig.rompath);
	sys_timer_sleep(5);
	sys_process_exit(0);
	}

	//Set Bios
	sprintf(Iniconfig.biospath,"%s/bios.bin",Iniconfig.biospath);

	printf("Run the emulator\n");

  while(is_running)
  {
	switch(state){
	case IN_MENU:
		RomBrowser();
	case IN_GAME:
	StateLoop();
	system_frame(0);
	Graphics->Draw(bitmap.viewport.w,bitmap.viewport.h,bitmap.data);
	SoundThread(0);
	Graphics->Swap();
	}
  }
	
	printf("done \n");
	
	 cellSysmoduleUnloadModule(CELL_SYSMODULE_IO);    
	 cellSysmoduleUnloadModule(CELL_SYSMODULE_FS); 
	 cellSysutilUnregisterCallback(0);  
    
	 return(-1);
}
